@@ -1,6 +1,4 @@
 import argparse
-import inspect
-import os
 import textwrap
 import logging
 from datetime import datetime
@@ -133,12 +131,6 @@ def run(game_selector: Union[str, Dict, GameSpec], model_selectors: List[backend
                     game_benchmark.filter_experiment.append(experiment_name)
                 time_start = datetime.now()
                 game_benchmark.run(player_models=player_models, results_dir=results_dir, task_selector=task_selector)
-                try:
-                    stdout_logger.info(f"Scoring game {game_spec['game_name']}")
-                    game_benchmark.compute_scores(results_dir)
-                except Exception as e:
-                    stdout_logger.info(f"There was a problem during scoring. See clembench.log for details.")
-                    logger.error(e, exc_info=True)
                 time_end = datetime.now()
                 logger.info(f'Running {game_spec["game_name"]} took {str(time_end - time_start)}')
         except Exception as e:
@@ -235,7 +227,7 @@ def cli(args: argparse.Namespace):
 
     To list available models: 
     $> clem list models
-    
+
     To list available backends: 
     $> clem list backends
 
